(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Xml
imports ProofChecker 
begin

datatype xml = XML string "(string \<times> string) list" "xml list" "string option" (* name, attributes, children, text *)
datatype xmldoc = XMLDOC "string" xml (* header, body *)

text {* An alternative induction scheme for xml. *}
lemma
 fixes P :: "xml \<Rightarrow> bool"
 assumes "\<And>n a cs t. (\<And>c. c \<in> set cs \<Longrightarrow> P c) \<Longrightarrow> P(XML n a cs t)"
 shows xml_induct[case_names XML, induct type: xml]: "P xml" and "\<And>xml. xml \<in> set xmls \<Longrightarrow> P xml"
using assms by (induct xml and xmls) auto

text {* Resolve problem with recursive dependencies in generated code. *}
lemma [code]:
  shows "HOL.eq (XML name1 attrs1 children1 text1) (XML name2 attrs2 children2 text2) \<longleftrightarrow>
    HOL.eq name1 name2 \<and> HOL.eq attrs1 attrs2 \<and> list_all2 HOL.eq children1 children2 \<and> HOL.eq text1 text2"
  by (simp_all add: eq list_all2_eq [symmetric])

abbreviation squote :: "char" where "squote \<equiv> Char Nibble2 Nibble7"
abbreviation tab :: "char" where "tab \<equiv> Char Nibble0 Nibble9"
abbreviation carriage_return :: "char" where "carriage_return \<equiv> Char Nibble0 NibbleD"
definition wspace :: "string" where "wspace = newline # tab # carriage_return # '' ''"
definition dquote :: "string" where "dquote = ''\"''"
definition letters :: "string" where "letters = ''abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_0123456789&;:''"

fun
  trim :: "string \<Rightarrow> string" 
where
  "trim [] = []" |
  "trim (c # s) = (if c mem wspace then trim s else c # s)"

fun
  splitter_fun :: "(char \<Rightarrow> bool) \<Rightarrow> string \<Rightarrow> string \<times> string"
where
  "splitter_fun _ [] = ([],[])" |
  "splitter_fun allowed (c # s) = (if allowed c
     then (let (l,r) = splitter_fun allowed s in (c # l, r))
     else ([], c # s))"

definition
  splitter :: "string \<Rightarrow> string \<Rightarrow> string \<times> string"
where
  "splitter allowed s = splitter_fun (\<lambda>c. c mem allowed) s"

types
  'a reply = "string + 'a"
  ('t, 'a) gen_parser = "'t list \<Rightarrow> ('a \<times> 't list) reply"
  'a parser = "(char, 'a) gen_parser"

subsection {* Setup parsers as monads *}

abbreviation (input)
  return :: "'a \<Rightarrow> ('t, 'a) gen_parser"
where
  "return x \<equiv> (\<lambda>ts. Error_Monad.return (x, ts))"

abbreviation (input)
  error :: "string \<Rightarrow> ('t, 'a) gen_parser"
where
  "error e \<equiv> (\<lambda>_. Error_Monad.error e)"

definition
  bind :: "('t, 'a) gen_parser \<Rightarrow> ('a \<Rightarrow> ('t, 'b) gen_parser) \<Rightarrow> ('t, 'b) gen_parser"
where
  "bind m f ts = do {
     (x, ts') \<leftarrow> m ts;
     f x ts'
   }"

setup {*
  Adhoc_Overloading.add_variant @{const_name bindM} @{const_name bind}
*}

(* types 'a endParser = "string \<Rightarrow> (string,'a)emonad" *)

definition
  update_tokens :: "('t list \<Rightarrow> 't list) \<Rightarrow> ('t, 't list) gen_parser"
where
  "update_tokens f ts \<equiv> Error_Monad.return (ts, f ts)"

definition
  get_tokens :: "('t, 't list) gen_parser"
where
  "get_tokens \<equiv> update_tokens id"

definition
  set_tokens :: "'t list \<Rightarrow> ('t, unit) gen_parser"
where
  "set_tokens ts \<equiv> update_tokens (\<lambda>_. ts) \<guillemotright> return ()"

(* TODO: move to Show *)
definition
  quote :: "shows \<Rightarrow> shows"
where
  "quote s \<equiv> shows_between (shows squote) s (shows squote)"

definition
  expecting :: "string \<Rightarrow> ('t::show, 'a) gen_parser"
where
  "expecting msg ts = Error_Monad.error
     (''expecting: '' @ msg @ '', but found: '' @ quote (shows (take 30 ts)) [])"

fun
  eoi :: "('t::show, unit) gen_parser"
where
  "eoi [] = Error_Monad.return ((), [])" |
  "eoi ts = expecting ''end of input'' ts"

fun
  oneof_aux :: "string list \<Rightarrow> string list \<Rightarrow> string parser"
where
  "oneof_aux allowed [] ts = expecting (''one of '' @ shows_list allowed []) ts" |
  "oneof_aux allowed (x # xs) ts = (if map snd (zip x ts) = x
     then Error_Monad.return (x, trim (List.drop (length x) ts))
     else oneof_aux allowed xs ts)"

definition
  oneof :: "string list \<Rightarrow> string parser"
where
  "oneof xs \<equiv> oneof_aux xs xs"

hide const oneof_aux

fun
  many :: "(char \<Rightarrow> bool) \<Rightarrow> (char list) parser"
where
  "many P [] = Error_Monad.return ([], [])" |
  "many P (t # ts) = (if P t
     then do {
       (rs, ts') \<leftarrow> many P ts;
       Error_Monad.return (t # rs, ts')
     } else Error_Monad.return ([], t # ts))"

definition
  manyof :: "char list \<Rightarrow> (char list) parser"
where
  "manyof cs \<equiv> many (\<lambda>c. c mem cs)"

definition
  spaces :: "unit parser"
where
  "spaces \<equiv> manyof wspace \<guillemotright> return ()"

fun
  scan_upto :: "string \<Rightarrow> string parser"
where
  "scan_upto end [] = Error_Monad.error (''did not find end-marker '' @ quote (shows end) [])" |
  "scan_upto end (t # ts) = (if map snd (zip end (t # ts)) = end
     then do {
       let l = length end;
       Error_Monad.return (end, List.drop l (t # ts))
     } else do {
       (res, ts') \<leftarrow> scan_upto end ts;
       Error_Monad.return (t # res, ts')
     })"

definition
  parse_name :: "string parser"
where
  "parse_name = do {
     n \<leftarrow> manyof letters;
     spaces;
     (if null n
       then error letters
       else return n)
   }"

definition
  parse_text :: "string option parser" 
where
  "parse_text = do {
     ts \<leftarrow> many (\<lambda>c. [c] \<noteq> ''<'');
     let text = trim ts;
     (if text = []
       then return None
       else return (Some (List.rev (trim (List.rev text)))))
   }"

definition
  parse_attribute_value :: "string parser"
where
  "parse_attribute_value = do {
     oneof [dquote];
     v \<leftarrow> many (\<lambda>c. [c] \<noteq> dquote);
     oneof [dquote];
     return v
   }"

function (sequential)
  parse_attributes :: "(string \<times> string) list parser"
where
  "parse_attributes [] = Error_Monad.return ([], [])" |
  "parse_attributes (c # s) = (if c mem ''/>''
     then Error_Monad.return ([], c # s)
     else ((do {
       k \<leftarrow> parse_name;
       oneof [''=''];
       v \<leftarrow> parse_attribute_value;
       atts \<leftarrow> parse_attributes;
       return ((k, v) # atts)
     }) (c # s)))"
by (pat_completeness) auto

function
  parse_nodes :: "xml list parser"
where
  "parse_nodes ts = (do {
     ts \<leftarrow> get_tokens;
     (if null ts \<or> take 2 ts = ''</''
       then return []
       else do {
         oneof [''<''];
         n \<leftarrow> parse_name;
         atts \<leftarrow> parse_attributes;
         e \<leftarrow> oneof [''/>'', ''>''];
         (if e = ''/>''
           then do {
             cs \<leftarrow> parse_nodes;
             return (XML n atts [] None # cs)
           } else do {
             t \<leftarrow> parse_text;
             cs \<leftarrow> parse_nodes;
             oneof [''</'' @ n @ ''>''];
             ns \<leftarrow> parse_nodes;
             return (XML n atts cs t # ns)
           })
       })
   }) ts"
by (pat_completeness) auto

definition
  parse_node :: "xml parser"
where
  "parse_node \<equiv> do {
     oneof [''<''];
     n \<leftarrow> parse_name;
     atts \<leftarrow> parse_attributes;
     e \<leftarrow> oneof [''/>'', ''>''];
     (if e = ''/>''
       then return (XML n atts [] None)
       else do {
         t \<leftarrow> parse_text;
         cs \<leftarrow> parse_nodes;
         oneof [''</'' @ n @ ''>''];
         return (XML n atts cs t)
       })
   }"

function
  parse_header :: "string parser"
where
  "parse_header ts = (do {
     ts \<leftarrow> get_tokens;
     (if take 2 (trim ts) = ''<?''
       then do {
         h \<leftarrow> scan_upto ''?>'';
         hs \<leftarrow> parse_header;
         return (h @ hs)
       } else do {
         spaces;
         return []
       })
   }) ts"
by (pat_completeness) auto

fun
  remove_comments_aux :: "bool \<Rightarrow> string \<Rightarrow> string"
where
  "remove_comments_aux False (c # cs) = (if (c = CHR ''<'' \<and> take 3 cs = ''!--'') 
     then remove_comments_aux True (tl cs)
     else c # remove_comments_aux False cs)" |
  "remove_comments_aux True (c # cs) = (if (c = CHR ''-'' \<and> take 2 cs = ''->'') 
     then remove_comments_aux False (drop 2 cs)
     else remove_comments_aux True cs)" |
  "remove_comments_aux _ [] = []"

definition
  remove_comments :: "string \<Rightarrow> string"
where
  "remove_comments \<equiv> remove_comments_aux False"

hide const remove_comments_aux

definition
  parse_doc :: "xmldoc parser"
where
  "parse_doc = debug ''0'' ''parsing txt to xml'' (do {
     update_tokens remove_comments;
     h \<leftarrow> parse_header;
     xml \<leftarrow> parse_node;
     eoi;
     return (XMLDOC h xml)
   })"

definition
  doc_of_string :: "string \<Rightarrow> string + xmldoc"
where
  "doc_of_string s = do {
     (doc, _) \<leftarrow> parse_doc s;
     Error_Monad.return doc
   }"


subsection {* Pretty Printing of XML nodes and documents *}

instantiation xml :: "show" begin
  definition
    shows_attr :: "string \<times> string \<Rightarrow> shows"
  where
    "shows_attr av = (shows (fst av) +@+ ''='' +#+ shows_string (dquote @ (snd av) @ dquote))"

  definition
    shows_attrs :: "(string\<times>string)list \<Rightarrow> shows"
  where
    "shows_attrs as = shows_map (\<lambda>a. '' '' +#+ shows_attr a) as"

  fun
    shows_XML_indent :: "string \<Rightarrow> nat \<Rightarrow> xml \<Rightarrow> shows"
  where
    "shows_XML_indent ind i (XML n a c t) = (
    ind +#+ ''<'' +#+ shows n +@+ shows_attrs a +@+ (
      if c = [] \<and> t = None then shows_string ''/>'' else (
        ''>'' +#+ (
          if c \<noteq> [] then (
            shows_map (shows_XML_indent (ind @ replicate i (CHR '' '')) i) c +@+ shows_string ind
          ) else id
        ) +@+ (if t = None then id else shows_string(the t)) +@+
          ''</'' +#+ shows n +@+ shows_string ''>''
      )
    )
  )"
  definition "shows_prec (d::nat) xml \<equiv> shows_XML_indent [newline] 2 xml"
  definition "shows_list (xs::xml list) \<equiv> shows_list_aux shows xs"
  
  lemma shows_attr_assoc: "(s +#+ shows_attr av) r @ t = (s +#+ shows_attr av) (r@t)"
    unfolding shows_attr_def by (cases av) (auto simp: show_defs o_def dquote_def)
  lemma shows_attr_assoc_elt: "\<forall>av\<in>set avs. (s +#+ shows_attr av) r @ t = (s +#+ shows_attr av) (r@t)"
    using shows_attr_assoc by best
  lemma shows_attrs_assoc: "shows_attrs as r @ s = shows_attrs as (r @ s)"
    unfolding shows_attrs_def shows_map_assoc[OF shows_attr_assoc_elt[of as "'' ''"]] by simp
  lemma assoc_xml': "shows_XML_indent ind i xml r @ s = shows_XML_indent ind i xml (r @ s)"
  proof (induct xml arbitrary: ind i r s)
    let ?elt = "shows_XML_indent"
    case (XML n a cs t ind i)
    hence elt: "\<And>r s.\<forall>c\<in>set cs. ?elt (ind @ replicate i CHR '' '') i c r @ s = ?elt (ind @ replicate i CHR '' '') i c (r@s)" by auto
    have IH: "\<And>r s. shows_map (?elt (ind @ replicate i CHR '' '') i) cs r @ s = shows_map (?elt (ind @ replicate i CHR '' '') i) cs (r@s)"
      by (rule shows_map_assoc[OF elt])
    show ?case
    proof (cases "cs = []")
      case True show ?thesis
      proof (cases t)
        case None
        with True have empty: "cs = [] \<and> t = None" by simp
        show ?thesis by (auto simp: shows_attrs_assoc if_P[OF empty])
      next
        case (Some ts)
        hence t: "\<not>(cs = [] \<and> t = None)" and some_t: "\<not> t = None" by auto
        from True have empty: "\<not> cs \<noteq> []" by simp
        show ?thesis by (auto simp: if_not_P[OF t] if_not_P[OF empty] if_not_P[OF some_t] shows_attrs_assoc)
      qed
    next
      case False
      hence ne: "\<not>(cs = [] \<and> t = None)" by simp
      show ?thesis
      proof (cases t)
        case None show ?thesis by (auto simp: if_not_P[OF ne] if_P[OF False] if_P[OF None] shows_attrs_assoc IH)
      next
        case (Some ts)
        hence some_t: "\<not> t = None" by simp
        show ?thesis by (auto simp: if_not_P[OF ne] if_P[OF False] if_not_P[OF some_t] shows_attrs_assoc IH)
      qed
    qed
  qed

  lemma assoc_xml: "shows_prec d (xml::xml) r @ s = shows_prec d xml (r @ s)" unfolding shows_prec_xml_def by (rule assoc_xml')
  lemma assoc_xml_elt: "\<forall>(x::xml)\<in>set xs. shows_prec d x r @ s = shows_prec d x (r @ s)" using assoc_xml by auto

  instance by
    (intro_classes,rule assoc_xml,unfold shows_list_xml_def)
    (rule shows_list_aux_assoc[OF assoc_xml_elt])
end

instantiation xmldoc :: "show" begin
  fun shows_xmldoc where "shows_xmldoc(XMLDOC h x) = (h +#+ shows_nl +@+ shows x)"
  definition "shows_prec (d::nat) doc = shows_xmldoc doc"
  definition "shows_list (xs::xmldoc list) \<equiv> shows_list_aux shows xs"

  lemma assoc_xmldoc: "shows_prec d (x::xmldoc) r @ s = shows_prec d x (r @ s)" by (cases x) (auto simp: shows_prec_xmldoc_def)
  lemma assoc_xmldoc_elt: "\<forall>(x::xmldoc)\<in>set xs. shows_prec d x r @ s = shows_prec d x (r @ s)" using assoc_xmldoc by auto

  instance by
    (intro_classes,rule assoc_xmldoc,unfold shows_list_xmldoc_def)
    (rule shows_list_aux_assoc[OF assoc_xmldoc_elt])
end


subsection {* XML Transformers *}
hide (open) const return
hide (open) const error

(* transformers on xml nodes *)
types
  'a xmlt = "xml \<Rightarrow> string + 'a"

definition
  map :: "'a xmlt \<Rightarrow> xml list \<Rightarrow> string + 'a list"
where
  "map = Error_Monad.mapm"
hide (open) const map

fun
  get_name :: "xml \<Rightarrow> string" 
where
  "get_name (XML name _ _ _) = name"
hide (open) const get_name

definition
  string2bool :: "string \<Rightarrow> string + bool"
where
  "string2bool s \<equiv> (if s = ''true''
     then Error_Monad.return True
     else (if s = ''false''
       then Error_Monad.return False
       else Error_Monad.error (''cannot convert '' @ s @ '' into Boolean'')))"

fun
  "text" :: "string \<Rightarrow> string xmlt"
where
  "text tag (XML n atts cs t) = 
     (if n = tag \<and> atts = [] \<and> cs = [] \<and> t \<noteq> None \<and> the t \<noteq> []
       then Error_Monad.return (the t)
       else Error_Monad.error (concat [''could not extract text for '', tag,'' from '',
         [newline], to_string (XML n atts cs t)]))"
hide (open) const "text"

fun
 bool :: "string \<Rightarrow> bool xmlt"
where
  "bool tag node = (Xml.text tag node \<guillemotright>= string2bool)"
hide (open) const bool

definition
  no_transform :: "string \<Rightarrow> xml \<Rightarrow> string + 'a"
where
  "no_transform tag xml = error (concat [''could not transform the following
     xml element (expected '', tag, '')'', [newline], to_string xml])"

fun
  xml0elements :: "string \<Rightarrow> 'a \<Rightarrow> 'a xmlt"
where
  "xml0elements tag x (XML name atts cs text) = 
     (if name = tag \<and> atts = [] \<and> cs = [] \<and> text = None
       then return x 
       else no_transform tag (XML name atts cs text))"

fun
  xml1element :: "string \<Rightarrow> 'a xmlt \<Rightarrow> ('a \<Rightarrow> 'b) \<Rightarrow> 'b xmlt"
where
  "xml1element tag p1 f (XML name atts cs text) = 
     (if name = tag \<and> atts = [] \<and> length cs = 1 \<and> text = None
       then do {
         x1 \<leftarrow> p1 (cs ! 0);
         return (f x1)
       } else no_transform tag (XML name atts cs text))"

fun
  xml2elements ::
    "string \<Rightarrow> 'a xmlt \<Rightarrow> 'b xmlt \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> 'c) \<Rightarrow> 'c xmlt"
where
  "xml2elements tag p1 p2 f (XML name atts cs text) = 
     (if name = tag \<and> atts = [] \<and> length cs = 2 \<and> text = None
       then do {
         x1 \<leftarrow> p1 (cs ! 0);
         x2 \<leftarrow> p2 (cs ! 1);
         return (f x1 x2)
       } else no_transform tag (XML name atts cs text))"

fun
  xml3elements ::
    "string \<Rightarrow> 'a xmlt \<Rightarrow> 'b xmlt \<Rightarrow> 'c xmlt \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> 'c \<Rightarrow> 'd) \<Rightarrow> 'd xmlt"
where
  "xml3elements tag p1 p2 p3 f (XML name atts cs text) = 
     (if name = tag \<and> atts = [] \<and> length cs = 3 \<and> text = None
       then do {
         x1 \<leftarrow> p1 (cs ! 0);
         x2 \<leftarrow> p2 (cs ! 1);
         x3 \<leftarrow> p3 (cs ! 2);
         return (f x1 x2 x3)
       } else no_transform tag (XML name atts cs text))"

fun
  xml4elements ::
    "string \<Rightarrow> 'a xmlt \<Rightarrow> 'b xmlt \<Rightarrow> 'c xmlt \<Rightarrow> 'd xmlt \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> 'c \<Rightarrow> 'd \<Rightarrow> 'e) \<Rightarrow> 'e xmlt"
where
  "xml4elements tag p1 p2 p3 p4 f (XML name atts cs text) = 
     (if name = tag \<and> atts = [] \<and> length cs = 4 \<and> text = None
       then do {
         x1 \<leftarrow> p1 (cs ! 0);
         x2 \<leftarrow> p2 (cs ! 1);
         x3 \<leftarrow> p3 (cs ! 2);
         x4 \<leftarrow> p4 (cs ! 3);
         return (f x1 x2 x3 x4)
       } else no_transform tag (XML name atts cs text))"

fun
  xml5elements ::
    "string \<Rightarrow> 'a xmlt \<Rightarrow> 'b xmlt \<Rightarrow> 'c xmlt \<Rightarrow> 'd xmlt \<Rightarrow> 'e xmlt
     \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> 'c \<Rightarrow> 'd \<Rightarrow> 'e \<Rightarrow> 'f) \<Rightarrow> 'f xmlt"
where
  "xml5elements tag p1 p2 p3 p4 p5 f (XML name atts cs text) =
     (if name = tag \<and> atts = [] \<and> length cs = 5 \<and> text = None
       then do {
         x1 \<leftarrow> p1 (cs ! 0);
         x2 \<leftarrow> p2 (cs ! 1);
         x3 \<leftarrow> p3 (cs ! 2);
         x4 \<leftarrow> p4 (cs ! 3);
         x5 \<leftarrow> p5 (cs ! 4);
         return (f x1 x2 x3 x4 x5)
       } else no_transform tag (XML name atts cs text))"

fun
  xml1to2elements :: "string \<Rightarrow> 'a xmlt \<Rightarrow> 'b xmlt \<Rightarrow> ('a \<Rightarrow> 'b option \<Rightarrow> 'c) \<Rightarrow> 'c xmlt"
where
  "xml1to2elements tag p1 p2 f (XML name atts cs text) = (
     let l = length cs in
     (if name = tag \<and> atts = [] \<and> l \<ge> 1 \<and> l \<le> 2 \<and> text = None
       then do {
         x1 \<leftarrow> p1 (cs ! 0);
         (if l = 2
           then do {
             x2 \<leftarrow> p2 (cs ! 1);
             return (f x1 (Some x2))
           } else return (f x1 None))
       } else no_transform tag (XML name atts cs text)))"

(* parses the first element, the second argument must be a boolean. If the boolean is true, then the last element is parsed additionally *)
fun
  xml2nd_choice :: "string \<Rightarrow> 'a xmlt \<Rightarrow> string \<Rightarrow> 'b xmlt \<Rightarrow> ('a \<Rightarrow> 'b option \<Rightarrow> 'c) \<Rightarrow> 'c xmlt"
where
  "xml2nd_choice tag p1 cn p2 f (XML name atts cs text) = (
     let l = length cs in
     (if name = tag \<and> atts = [] \<and> l \<ge> 2 \<and> text = None
       then do {
         x1 \<leftarrow> p1 (cs ! 0);
         b \<leftarrow> Xml.bool cn (cs ! 1);
         (if b
           then do {
             x2 \<leftarrow> p2 (cs ! (l - 1));
             return (f x1 (Some x2))
           } else return (f x1 None))
       } else no_transform tag (XML name atts cs text)))"

fun
  xml2to3elements :: "string \<Rightarrow> 'a xmlt \<Rightarrow> 'b xmlt \<Rightarrow> 'c xmlt \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> 'c option \<Rightarrow> 'd) \<Rightarrow> 'd xmlt"
where
  "xml2to3elements tag p1 p2 p3 f (XML name atts cs text) = (
     let l = length cs in
     (if name = tag \<and> atts = [] \<and> l \<ge> 2 \<and> l \<le> 3 \<and> text = None
       then do {
         x1 \<leftarrow> p1 (cs ! 0);
         x2 \<leftarrow> p2 (cs ! 1);
         (if l = 3
           then do {
             x3 \<leftarrow> p3 (cs ! 2);
             return (f x1 x2 (Some x3))
           } else return (f x1 x2 None))
       } else no_transform tag (XML name atts cs text)))"

fun
  xml3to4elements :: "string \<Rightarrow> 'a xmlt \<Rightarrow> 'b xmlt \<Rightarrow> 'c xmlt \<Rightarrow> 'd xmlt \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> 'c option \<Rightarrow> 'd \<Rightarrow> 'e) \<Rightarrow> 'e xmlt"
where
  "xml3to4elements tag p1 p2 p3 p4 f (XML name atts cs text) = (
     let l = length cs in
     (if name = tag \<and> atts = [] \<and> l \<ge> 3 \<and> l \<le> 4 \<and> text = None
       then do {
         x1 \<leftarrow> p1 (cs ! 0);
         x2 \<leftarrow> p2 (cs ! 1);
         (if l = 4
           then do {
             x3 \<leftarrow> p3 (cs ! 2);
             x4 \<leftarrow> p4 (cs ! 3);
             return (f x1 x2 (Some x3) x4)
           } else do {
             x4 \<leftarrow> p4 (cs ! 2);
             return (f x1 x2 None x4)
           } )
       } else no_transform tag (XML name atts cs text)))"

fun
  xml_many_elements :: "string \<Rightarrow> 'a xmlt \<Rightarrow> ('a list \<Rightarrow> 'b) \<Rightarrow> 'b xmlt"
where
  "xml_many_elements tag p f (XML name atts cs text) =
     (if name = tag \<and> atts = [] \<and> text = None
       then (Xml.map p cs \<guillemotright>= (return \<circ> f))
       else no_transform tag (XML name atts cs text))"

fun
  xml1many_elements_gen :: "string \<Rightarrow> 'a xmlt \<Rightarrow> ('a \<Rightarrow> 'b xmlt) \<Rightarrow> ('a \<Rightarrow> 'b list \<Rightarrow> 'c) \<Rightarrow> 'c xmlt"
where
  "xml1many_elements_gen tag p1 p2 f (XML name atts cs text) =
     (if name = tag \<and> atts = [] \<and> text = None \<and> cs \<noteq> []
       then do {
         x \<leftarrow> p1 (hd cs);
         xs \<leftarrow> Xml.map (p2 x) (tl cs);
         return (f x xs)
       } else no_transform tag (XML name atts cs text))"

definition
  xml1many_elements :: "string \<Rightarrow> 'a xmlt \<Rightarrow> 'b xmlt \<Rightarrow> ('a \<Rightarrow> 'b list \<Rightarrow> 'c) \<Rightarrow> 'c xmlt"
where
  "xml1many_elements tag p1 p2 = xml1many_elements_gen tag p1 (\<lambda>_. p2)"

fun
  xml2many_elements :: "string \<Rightarrow> 'a xmlt \<Rightarrow> 'b xmlt \<Rightarrow> 'c xmlt \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> 'c list \<Rightarrow> 'd) \<Rightarrow> 'd xmlt"
where
  "xml2many_elements tag p1 p2 p3 f (XML name atts cs text) =
     (if name = tag \<and> atts = [] \<and> text = None \<and> length cs \<ge> 2
       then do {
         x \<leftarrow> p1 (cs ! 0);
         y \<leftarrow> p2 (cs ! 1);
         xs \<leftarrow> Xml.map p3 (tl (tl cs));
         return (f x y xs)
       } else no_transform tag (XML name atts cs text))"

fun
  xml1many2elements_gen :: "string \<Rightarrow> 'a xmlt \<Rightarrow> ('a \<Rightarrow> 'b xmlt) \<Rightarrow> 'c xmlt \<Rightarrow> 'd xmlt \<Rightarrow> ('a \<Rightarrow> 'b list \<Rightarrow> 'c \<Rightarrow> 'd \<Rightarrow> 'e) \<Rightarrow> 'e xmlt"
where
  "xml1many2elements_gen tag p1 p2 p3 p4 f (XML name atts cs text) = (
     let ds = List.rev cs; l = length cs in
     (if name = tag \<and> atts = [] \<and> text = None \<and> l \<ge> 3
       then do {
         x \<leftarrow> p1 (cs ! 0);
         xs \<leftarrow> Xml.map (p2 x) (tl (take (l - 2) cs));
         y \<leftarrow> p3 (ds ! 1);
         z \<leftarrow> p4 (ds ! 0);
         return (f x xs y z)
       } else no_transform tag (XML name atts cs text)))"

fun
  xml1many2elements :: "string \<Rightarrow> 'a xmlt \<Rightarrow> 'b xmlt \<Rightarrow> 'c xmlt \<Rightarrow> 'd xmlt \<Rightarrow> ('a \<Rightarrow> 'b list \<Rightarrow> 'c \<Rightarrow> 'd \<Rightarrow> 'e) \<Rightarrow> 'e xmlt"
where
  "xml1many2elements tag p1 p2 = xml1many2elements_gen tag p1 (\<lambda>_. p2)"

fun
  xml_many2elements :: "string \<Rightarrow> 'a xmlt \<Rightarrow> 'b xmlt \<Rightarrow> 'c xmlt \<Rightarrow> ('a list \<Rightarrow> 'b \<Rightarrow> 'c \<Rightarrow> 'd) \<Rightarrow> 'd xmlt"
where
  "xml_many2elements tag p1 p2 p3 f (XML name atts cs text) = (
     let ds = List.rev cs in
     (if name = tag \<and> atts = [] \<and> text = None \<and> length cs \<ge> 2
       then do {
         xs \<leftarrow> Xml.map p1 (List.rev (tl (tl ds)));
         y \<leftarrow> p2 (ds ! 1);
         z \<leftarrow> p3 (ds ! 0);
         return (f xs y z)
       } else no_transform tag (XML name atts cs text)))"

fun
  xml_options :: "(string \<times> 'a xmlt) list \<Rightarrow> 'a xmlt"
where
  "xml_options ps x =
     (case lookup (Xml.get_name x) ps of
       None \<Rightarrow> error (concat [''expected one of: '', concat (map (\<lambda>p. fst p @ '' '') ps), [newline], ''but found'', [newline], to_string x])
     | Some p \<Rightarrow> p x)"

fun
  xml_change :: "'a xmlt \<Rightarrow> ('a \<Rightarrow> 'b) \<Rightarrow> 'b xmlt"
where
  "xml_change p f x = (p x \<guillemotright>= (\<lambda>a. return (f a)))"

fun
  digit2int :: "char \<Rightarrow> string + int"
where
  "digit2int x =
     (if x = CHR ''0'' then return 0 else
     (if x = CHR ''1'' then return 1 else
     (if x = CHR ''2'' then return 2 else
     (if x = CHR ''3'' then return 3 else
     (if x = CHR ''4'' then return 4 else
     (if x = CHR ''5'' then return 5 else
     (if x = CHR ''6'' then return 6 else
     (if x = CHR ''7'' then return 7 else
     (if x = CHR ''8'' then return 8 else
     (if x = CHR ''9'' then return 9 else 
       error ([x] @ '' is not a digit'')))))))))))"

fun
  string2int_aux :: "int \<Rightarrow> string \<Rightarrow> string + int" 
where
  "string2int_aux n [] = return n" |
  "string2int_aux n (d # s) = (digit2int d \<guillemotright>= (\<lambda>m. string2int_aux (10 * n + m) s))"

definition
  string2int :: "string \<Rightarrow> string + int"
where
  "string2int s =
     (if s = []
       then error ''cannot convert empty string into number'' 
       else if take 1 s = ''-''
         then string2int_aux 0 (tl s) \<guillemotright>= (\<lambda> i. return (0 - i))
         else string2int_aux 0 s)"

hide const string2int_aux

fun
  int :: "string \<Rightarrow> int xmlt"
where
  "int tag x = (Xml.text tag x \<guillemotright>= string2int)"
hide (open) const int

fun
  nat :: "string \<Rightarrow> nat xmlt"
where
  "nat tag x = (Xml.text tag x \<guillemotright>= string2int \<guillemotright>= (\<lambda>i. return (Int.nat i)))"
hide (open) const nat

definition
  rat :: "rat xmlt"
where
  "rat \<equiv> xml_options [(''integer'', xml_change (Xml.int ''integer'') (\<lambda>x. Fract x 1)),
        (''rational'', xml2elements ''rational'' (Xml.int ''numerator'') (Xml.int ''denominator'') Fract)]"
hide (open) const rat

fun
  var :: "('a, string) term xmlt"
where
  "var x = (Xml.text ''var'' x \<guillemotright>= (return \<circ> Var))"
hide (open) const var

function
  "term" :: "'a xmlt \<Rightarrow> ('a, string) term xmlt"
where
  "term xml2name x = 
     xml_options [(''var'', Xml.var), (''funapp'', xml1many_elements ''funapp'' xml2name (xml1element ''arg'' (term xml2name) id) Fun)] x"  
by (pat_completeness) auto
hide (open) const "term"

function
  ctxt :: "'a xmlt \<Rightarrow> ('a, string) ctxt xmlt"
where
  "ctxt xml2name x = xml_options [
     (''box'', xml0elements ''box'' Hole),  
     (''funContext'', xml4elements ''funContext'' 
       xml2name 
       (xml_many_elements ''before'' (Xml.term xml2name) id)
       (ctxt xml2name)
       (xml_many_elements ''after'' (Xml.term xml2name) id)
       More)
   ] x"  
by (pat_completeness) auto
hide (open) const ctxt

definition
  flat_contexts :: "'a xmlt \<Rightarrow> ('a, string) ctxt list xmlt"
where
  "flat_contexts xml2name = xml_many_elements ''flatContexts'' (Xml.ctxt xml2name) id"
hide (open) const flat_contexts

definition
  sub :: "'a xmlt \<Rightarrow> ('a, string) subL xmlt"
where
  "sub xml2name = 
     xml_many_elements ''substitution''
       (xml2elements ''substEntry'' 
         (Xml.text ''var'')
         (Xml.term xml2name)
         Pair)
       id"
hide (open) const sub

primrec
  list2position :: "nat list \<Rightarrow> pos"
where
  "list2position [] = \<epsilon>" |
  "list2position (n # ns) = n <# list2position ns"

definition
  position :: "nat xmlt"
where
  "position \<equiv> xml_change (Xml.nat ''position'') (\<lambda>n. n - 1)"
hide (open) const position

definition
  pos :: "pos xmlt"
where
  "pos \<equiv> xml_many_elements ''positionInTerm'' Xml.position list2position"
hide (open) const pos

fun
  sp :: "'a xmlt \<Rightarrow> 'a spL xmlt"
where
  "sp xml2name = xml_many_elements ''argumentFilter''
     (xml3elements ''argumentFilterEntry''
       xml2name
       (Xml.nat ''arity'')
       (Xml.nat ''collapsing'')
       (\<lambda>f a n. Pair f (n - 1)))
     SimpProj"
hide (open) const sp

definition
  rule :: "'a xmlt \<Rightarrow> ('a, string) rule xmlt"
where
  "rule xml2name = xml2elements ''rule'' (xml1element ''lhs'' (Xml.term xml2name) id) (xml1element ''rhs'' (Xml.term xml2name) id) Pair"
hide (open) const rule 

definition
  rstep :: "'a xmlt \<Rightarrow> (pos \<times> ('a, string) rule \<times> bool \<times> ('a, string) term) xmlt"
where
  "rstep xml2name = xml3to4elements ''rewriteStep'' Xml.pos (Xml.rule xml2name) (xml0elements ''relative'' ()) (Xml.term xml2name) (\<lambda>p r rel t. (p, r, rel = None, t))"
hide (open) const rstep

definition
  rsteps :: "'a xmlt \<Rightarrow> (('a,string)term \<times> ('a,string)prseq) xmlt"
where "rsteps xml2name \<equiv> xml1many_elements ''rewriteSequence'' 
             (xml1element ''startTerm'' (Xml.term xml2name) id)
             (Xml.rstep xml2name)
             Pair"
hide (open) const rsteps

definition 
  loop :: "'a xmlt \<Rightarrow> (('a,string)term \<times> ('a,string)prseq \<times> ('a,string)subL \<times> ('a,string)ctxt) xmlt"
where "loop xml2name \<equiv> xml3elements ''loop''
             (Xml.rsteps xml2name)
             (Xml.sub xml2name)
             (Xml.ctxt xml2name)
             (\<lambda> (s,rseq) \<sigma> c. (s,rseq, \<sigma>, c))"
hide (open) const loop

definition
  start_term :: "'a xmlt \<Rightarrow> ('a, string) term \<Rightarrow> ('a, string) term xmlt"
where
  "start_term xml2name t = xml1element ''startTerm'' (\<lambda>xml. (
     Xml.term xml2name xml \<guillemotright>= (\<lambda>s. if s = t
       then return t
       else error (''<startTerm> does not match lhs'')
     )
   )) id"
hide (open) const start_term

definition
  rseq :: "'a xmlt \<Rightarrow> 'a sp \<Rightarrow> ('a, string) rule \<Rightarrow> (('a, string) rule \<times> (pos \<times> ('a, string) rule \<times> ('a, string) term) list) xmlt"
where
  "rseq xml2name \<pi> r = xml1many_elements ''rewriteSequence''
     (Xml.start_term xml2name (proj_term \<pi> (fst r))) (Xml.rstep xml2name) (\<lambda>_ rseq. (r, map (\<lambda> (x,y,_,z). (x,y,z)) rseq))"
hide (open) const rseq

fun
  projected_rseq :: "'a xmlt \<Rightarrow> 'a sp \<Rightarrow> (('a, string) rule \<times> (pos \<times> ('a, string) rule \<times> ('a, string) term) list) xmlt"
where
  "projected_rseq xml2name \<pi> (XML name atts cs text) = (
     let tag = ''projectedRewriteSequence'' in (if name = tag \<and> atts = [] \<and> text = None \<and> length cs = 2
       then (Xml.rule xml2name (cs ! 0) \<guillemotright>= (\<lambda>r. Xml.rseq xml2name \<pi> r (cs ! 1)))
       else no_transform tag (XML name atts cs text)
     )
   )"
hide (open) const projected_rseq

definition
  rules :: "'a xmlt \<Rightarrow> ('a,string) rules xmlt"
where
  "rules xml2name = xml_many_elements ''rules'' (Xml.rule xml2name) id"
hide (open) const rules

definition
  funcsym :: "(string \<times> nat) xmlt"
where
  "funcsym = xml2elements ''funcsym'' (Xml.text ''name'') (Xml.nat ''arity'') Pair"
hide (open) const funcsym 

definition
  "signature" :: "(string \<times> nat) list xmlt"
where
  "signature = xml_many_elements ''signature'' Xml.funcsym id"
hide (open) const "signature"

types labelType = "nat list" 
types lsymbol = "(string, labelType) lab"
types llsymbol = "(lsymbol shp, labelType) lab"

definition
  label :: "'a xmlt \<Rightarrow> (labelType + 'a list) xmlt"
where
  "label xml2name \<equiv> xml_options [
     (''numberLabel'', xml_many_elements ''numberLabel'' (Xml.nat ''number'') Inl),
     (''symbolLabel'', xml_many_elements ''symbolLabel'' xml2name Inr)
   ]"
hide (open) const label 

function
  plain_name :: "lsymbol xmlt"
where
  "plain_name x = xml_options [
     (''name'', xml_change (Xml.text ''name'') UnLab),
     (''labeledSymbol'', xml2elements ''labeledSymbol'' plain_name (Xml.label plain_name)
     (\<lambda>f l. case l of
       Inl ls  \<Rightarrow> Lab f ls
     | Inr ls \<Rightarrow> FunLab f ls))
   ] x"
by pat_completeness auto
hide (open) const plain_name

definition
  sharp_lab_name :: "lsymbol shp xmlt"
where
  "sharp_lab_name x = (if Xml.get_name x = ''sharp''
     then xml1element ''sharp'' Xml.plain_name Sharp x
     else xml_change Xml.plain_name Plain x)"
hide (open) const sharp_lab_name

definition
  sharp_name :: "string shp xmlt"
where
  "sharp_name x = (if Xml.get_name x = ''sharp''
     then xml1element ''sharp'' (Xml.text ''name'') Sharp x
     else xml_change (Xml.text ''name'') Plain x)"
hide (open) const sharp_name

fun
  lab_sharp_name :: "nat \<Rightarrow> llsymbol xmlt"
where
  "lab_sharp_name 0 = xml_change Xml.sharp_lab_name UnLab" |
  "lab_sharp_name (Suc n) = xml2elements ''labeledSymbol'' (lab_sharp_name n) (Xml.label (lab_sharp_name n))
     (\<lambda>f l. case l of
       Inl ls \<Rightarrow> Lab f ls
     | Inr ls \<Rightarrow> FunLab f ls)"
hide (open) const lab_sharp_name

fun
  fit_length :: "'a \<Rightarrow> nat \<Rightarrow> 'a list \<Rightarrow> 'a list"
where
  "fit_length _ 0 _ = []" |
  "fit_length a n [] = replicate n a" |
  "fit_length a (Suc n) (b # as) = b # (fit_length a n as)"

function
  arith_fun :: "arithFun xmlt"
where
  "arith_fun xml = xml1element ''arithFunction'' (xml_options [ 
     (''natural'', xml_change (Xml.nat ''natural'') Const),
     (''variable'', xml_change (Xml.nat ''variable'') (\<lambda>n::nat. Arg (n - 1))),
     (''sum'', xml_many_elements ''sum'' arith_fun Sum),
     (''product'', xml_many_elements ''product'' arith_fun Prod),
     (''min'', xml_many_elements ''min'' arith_fun Min),
     (''max'', xml_many_elements ''max'' arith_fun Max)
   ]) id  xml"
by (pat_completeness, auto)
hide (open) const arith_fun

definition
  sl_inter :: "('f, 'l) lab xmlt \<Rightarrow> ('f, 'l) sl_inter option xmlt"
where
  "sl_inter xml2name \<equiv> xml1element ''model'' (xml_options [
     (''finiteModel'', xml1many_elements ''finiteModel''
       (xml_change (Xml.nat ''carrierSize'') (\<lambda> n :: nat. n - 1))
         (xml3elements ''interpret''
           xml2name
           (Xml.nat ''arity'')
           Xml.arith_fun
           (\<lambda>f a inter. ((f, a), inter)))
           (\<lambda>n xs. Some (SL_Inter n xs))),
     (''rootLabeling'',xml0elements ''rootLabeling'' None)
   ]) id"
hide (open) const sl_inter

datatype 'a poly = PVar nat | PNum 'a | PSum "'a poly list" | PMult "'a poly list"

fun
  sum_lpoly :: "('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> 'a list"
where
  "sum_lpoly _ [] ys = ys" |
  "sum_lpoly _ xs [] = xs" |
  "sum_lpoly pl (x # xs) (y # ys) = pl x y # sum_lpoly pl xs ys"

fun
  poly_to_lpoly :: "'a \<Rightarrow> 'a \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> 'a poly \<Rightarrow> (string + 'a \<times> 'a list)"
where "poly_to_lpoly _ _ _ _ (PNum i) = return (i,[])"
 | "poly_to_lpoly ze on pl ti (PVar x) = return (ze, replicate x ze @ [on])"
 | "poly_to_lpoly ze on pl ti (PSum [p]) = poly_to_lpoly ze on pl ti p"
 | "poly_to_lpoly ze on pl ti (PSum (p # q # ps)) = (
      poly_to_lpoly ze on pl ti p 
  \<guillemotright>= (\<lambda> (cp,ncp). poly_to_lpoly ze on pl ti (PSum (q # ps))
  \<guillemotright>= (\<lambda> (cq,ncq). return (pl cp cq,sum_lpoly pl ncp ncq))))"
 | "poly_to_lpoly ze on pl ti (PMult [p]) = poly_to_lpoly ze on pl ti p"
 | "poly_to_lpoly ze on pl ti (PMult (p # q # ps)) = (
      poly_to_lpoly ze on pl ti p
  \<guillemotright>= (\<lambda> (cp,ncp). poly_to_lpoly ze on pl ti (PMult (q # ps))
  \<guillemotright>= (\<lambda> (cq,ncq). if (forall (op = ze) ncp) then return (ti cp cq, map (ti cp) ncq)
             else (if (forall (op = ze) ncq) then return (ti cp cq, map (\<lambda> x. ti x cq) ncp)
             else error ''non-linear polynomials not supported''))))"

function
  polynomial :: "'a xmlt \<Rightarrow> 'a poly xmlt"
where
  "polynomial xml2coeff x = xml1element ''polynomial'' 
     (xml_options [
       (''coefficient'',xml1element ''coefficient'' xml2coeff PNum), 
       (''variable'', xml_change (Xml.nat ''variable'') (\<lambda>n. PVar (n - 1))),
       (''sum'', xml_many_elements ''sum'' (polynomial xml2coeff) PSum),
       (''product'', xml_many_elements ''product'' (polynomial xml2coeff) PMult)
     ]) id x"
by pat_completeness auto
hide (open) const polynomial

definition
  int_coeff :: "int xmlt"
where
  "int_coeff \<equiv> Xml.int ''integer''"
hide (open) const int_coeff

definition
  vec_coeff :: "'a xmlt \<Rightarrow> 'a vec xmlt"
where
  "vec_coeff xml2coeff \<equiv> xml_many_elements ''vector''
     (xml1element ''coefficient'' xml2coeff id) id"
hide (open) const vec_coeff

(* allow vectors and convert them to matrices by filling in zeros *)
definition
  mat_coeff :: "'a \<Rightarrow> 'a xmlt \<Rightarrow> 'a mat xmlt"
where
  "mat_coeff ze xml2coeff \<equiv> xml_options [
     (''matrix'', xml_many_elements ''matrix'' (Xml.vec_coeff xml2coeff) id),
     (''vector'', xml_change (Xml.vec_coeff xml2coeff) (\<lambda>v. v # replicate (length v - 1) (replicate (length v) ze)))
   ]"
hide (open) const mat_coeff

definition
  arctic_coeff :: "arctic xmlt"
where
  "arctic_coeff \<equiv> xml_options [
     (''integer'', xml_change (Xml.int ''integer'') Num_arc),
     (''minusInfinity'', xml0elements ''minusInfinity'' MinInfty)
   ]"
hide (open) const arctic_coeff

definition
  arctic_rat_coeff :: "arctic_rat xmlt"
where
  "arctic_rat_coeff \<equiv> xml_options [
     (''integer'', xml_change Xml.rat Num_arc_rat),
     (''rational'', xml_change Xml.rat Num_arc_rat),
     (''minusInfinity'', xml0elements ''minusInfinity'' MinInfty_rat)
   ]"
hide (open) const arctic_rat_coeff


datatype Domain = Natural | Arctic | Arctic_rat rat | Int_mat nat nat | Arc_mat nat | Arc_rat_mat rat nat | Rational rat | Rat_mat rat nat nat

definition
  rat_domain :: "rat xmlt"
where
  "rat_domain \<equiv> xml1element ''rationals'' (xml1element ''delta'' Xml.rat id) id"
hide (open) const rat_domain

definition
  basic_domain :: "Domain xmlt"
where
  "basic_domain \<equiv> xml_options [
     (''naturals'', xml0elements ''naturals'' Natural),
     (''arctic'', xml1element ''arctic'' 
       (xml1element ''domain'' (xml_options [
         (''naturals'', xml0elements ''naturals'' Arctic),
         (''integers'', xml0elements ''integers'' Arctic),
         (''rationals'', xml_change Xml.rat_domain (\<lambda> r. Arctic_rat r))
       ]) id) id),
     (''rationals'', xml_change Xml.rat_domain Rational)]"
hide (open) const basic_domain

definition
  interpretation_type :: "Domain xmlt"
where
  "interpretation_type \<equiv> (xml1element ''type'' 
     (xml_options [
       (''polynomial'', (xml2elements ''polynomial'' 
       (xml1element ''domain'' 
       (xml_options [
         (''naturals'', Xml.basic_domain),
         (''arctic'', Xml.basic_domain),
         (''rationals'', Xml.basic_domain),
         (''matrices'',xml3elements ''matrices''
           (Xml.nat ''dimension'') 
           (Xml.nat ''strictDimension'') 
           (xml1element ''domain'' Xml.basic_domain id)
           (\<lambda> di sd domain. case domain of
             Natural \<Rightarrow> Int_mat di sd
           | Rational delta  \<Rightarrow> Rat_mat delta di sd
           | Arctic \<Rightarrow> Arc_mat di
           | Arctic_rat delta \<Rightarrow> Arc_rat_mat delta di
         ))
       ]) 
     id)
     (Xml.nat ''degree'') 
     (\<lambda>type _ . type))),
     (''matrixInterpretation'',xml3elements ''matrixInterpretation''
       (xml1element ''domain'' Xml.basic_domain id)
       (Xml.nat ''dimension'') 
       (Xml.nat ''strictDimension'') 
       (\<lambda>domain di sd. case domain of
         Natural \<Rightarrow> Int_mat di sd
       | Rational delta  \<Rightarrow> Rat_mat delta di sd
       | Arctic \<Rightarrow> Arc_mat di
       | Arctic_rat delta \<Rightarrow> Arc_rat_mat delta di
     ))
   ])
 id)"
hide (open) const interpretation_type

datatype ('a,'b,'c,'d,'e,'f,'g,'h)alter = Alter1 'a | Alter2 'b | Alter3 'c | Alter4 'd | Alter5 'e | Alter6 'f | Alter7 'g | Alter8 'h
fun alter1 where "alter1 (Alter1 x) = x"
fun alter2 where "alter2 (Alter2 x) = x"
fun alter3 where "alter3 (Alter3 x) = x"
fun alter4 where "alter4 (Alter4 x) = x"
fun alter5 where "alter5 (Alter5 x) = x"
fun alter6 where "alter6 (Alter6 x) = x"
fun alter7 where "alter7 (Alter7 x) = x"
fun alter8 where "alter8 (Alter8 x) = x"

definition
  "interpretation" :: "'a xmlt \<Rightarrow> 'a redpair xmlt"
where
  "interpretation xml2name \<equiv> 
      (xml1many_elements_gen ''interpretation'' 
       Xml.interpretation_type
       (\<lambda>type. case type of
          Natural \<Rightarrow> (\<lambda>c. (xml3elements ''interpret''
                        xml2name
                        (Xml.nat ''arity'')
                        (\<lambda>x. Xml.polynomial Xml.int_coeff x \<guillemotright>= poly_to_lpoly (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c))
                        (\<lambda>f a poly.  (Alter1 ((f,a),(fst poly, fit_length (get_comp_0 c) a (snd poly))))))) 
                    int_components
        | Rational d \<Rightarrow> (\<lambda> c. (xml3elements ''interpret''
                        xml2name
                        (Xml.nat ''arity'')
                        (\<lambda>x. Xml.polynomial Xml.rat x \<guillemotright>= poly_to_lpoly (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c))
                        (\<lambda>f a poly.  (Alter2 ((f,a),(fst poly, fit_length (get_comp_0 c) a (snd poly))))))) 
                    (rat_components d)
        | Arctic \<Rightarrow> (\<lambda> c. (xml3elements ''interpret''
                        xml2name
                        (Xml.nat ''arity'')
                        (\<lambda>x. Xml.polynomial Xml.arctic_coeff x \<guillemotright>= poly_to_lpoly (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c))
                        (\<lambda>f a poly.  (Alter3 ((f,a),(fst poly, fit_length (get_comp_0 c) a (snd poly))))))) 
                    arctic_components
        | Arctic_rat d \<Rightarrow> (\<lambda> c. (xml3elements ''interpret''
                        xml2name
                        (Xml.nat ''arity'')
                        (\<lambda>x. Xml.polynomial Xml.arctic_rat_coeff x \<guillemotright>= poly_to_lpoly (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c))
                        (\<lambda>f a poly.  (Alter7 ((f,a),(fst poly, fit_length (get_comp_0 c) a (snd poly))))))) 
                    (arctic_rat_components d)
        | Int_mat n sd \<Rightarrow> (\<lambda> c. (xml3elements ''interpret''
                        xml2name
                        (Xml.nat ''arity'')
                        (\<lambda>x. Xml.polynomial (Xml.mat_coeff 0 Xml.int_coeff) x \<guillemotright>= poly_to_lpoly (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c))
                        (\<lambda>f a poly.  (Alter4 ((f,a),(fst poly, fit_length (get_comp_0 c) a (snd poly))))))) 
                    (mat_components n sd int_components)
        | Rat_mat d n sd \<Rightarrow> (\<lambda>c. (xml3elements ''interpret''
                        xml2name
                        (Xml.nat ''arity'')
                        (\<lambda>x. Xml.polynomial (Xml.mat_coeff 0 Xml.rat) x \<guillemotright>= poly_to_lpoly (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c))
                        (\<lambda>f a poly.  (Alter5 ((f,a),(fst poly, fit_length (get_comp_0 c) a (snd poly))))))) 
                    (mat_components n sd (rat_components d))
        | Arc_mat n \<Rightarrow> (\<lambda>c. (xml3elements ''interpret''
                        xml2name
                        (Xml.nat ''arity'')
                        (\<lambda>x. Xml.polynomial (Xml.mat_coeff MinInfty Xml.arctic_coeff) x \<guillemotright>= poly_to_lpoly (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c))
                        (\<lambda> f a poly. (Alter6 ((f,a),(fst poly, fit_length (get_comp_0 c) a (snd poly))))))) 
                    (mat_arctic_components n arctic_components)
        | Arc_rat_mat d n \<Rightarrow> (\<lambda>c. (xml3elements ''interpret''
                        xml2name
                        (Xml.nat ''arity'')
                        (\<lambda>x. Xml.polynomial (Xml.mat_coeff MinInfty_rat Xml.arctic_rat_coeff) x \<guillemotright>= poly_to_lpoly (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c))
                        (\<lambda>f a poly. (Alter8 ((f,a),(fst poly, fit_length (get_comp_0 c) a (snd poly))))))) 
                    (mat_arctic_components n (arctic_rat_components d))
        )
      (\<lambda> type pi. case type of 
            Natural \<Rightarrow> Int_carrier (map alter1 pi)
          | Rational d \<Rightarrow> Rat_carrier d (map alter2 pi)
          | Arctic \<Rightarrow> Arctic_carrier (map alter3 pi)
          | Arctic_rat d \<Rightarrow> Arctic_rat_carrier d (map alter7 pi)
          | Int_mat n sd \<Rightarrow> Int_mat_carrier n sd (map alter4 pi)
          | Rat_mat d n sd \<Rightarrow> Rat_mat_carrier d n sd (map alter5 pi)
          | Arc_mat n \<Rightarrow> Arc_mat_carrier n (map alter6 pi)
          | Arc_rat_mat d n \<Rightarrow> Arc_rat_mat_carrier d n (map alter8 pi)
      ))
    "
hide (open) const "interpretation"

definition
  afs :: "'a xmlt \<Rightarrow> 'a afs_list xmlt"
where
  "afs xml2name \<equiv> xml_many_elements ''argumentFilter'' 
  (xml3elements ''argumentFilterEntry''
      xml2name
      (Xml.nat ''arity'')
      (xml_options [(''collapsing'',xml_change (Xml.nat ''collapsing'') (\<lambda> i. Collapse (i - 1))),
                   (''nonCollapsing'', xml_many_elements ''nonCollapsing'' Xml.position AFList)])
      (\<lambda> f a e. ((f,a), e))
  ) id"
hide (open) const afs

definition
  precedence :: "'a xmlt \<Rightarrow> 'a lpo_repr xmlt"
where
  "precedence xml2name \<equiv> xml_many_elements ''statusPrecedence'' 
  (xml4elements ''statusPrecedenceEntry''
      xml2name
      (Xml.nat ''arity'')
      (Xml.nat ''precedence'')
      (xml_options [(''lex'', xml0elements ''lex'' ()), (''mul'', \<lambda> x. error ''multiset-comparisons not supported'')])
      (\<lambda>f a p s. ((f,a), p))
  ) id"
hide (open) const precedence

definition
  path_order :: "'a xmlt \<Rightarrow> 'a redpair xmlt"
where
  "path_order xml2name \<equiv> xml1to2elements ''pathOrder''
   (Xml.precedence xml2name)
   (Xml.afs xml2name)
   (\<lambda>pr afo. case afo of None \<Rightarrow> LPO pr [] | Some af \<Rightarrow> LPO pr af)"
hide (open) const path_order

definition
  redpair :: "'a xmlt \<Rightarrow> 'a redpair xmlt"
where
  "redpair xml2name = 
     xml1element ''redPair'' 
     (xml_options [(''interpretation'', Xml.interpretation xml2name), (''pathOrder'', Xml.path_order xml2name)]) id"
hide (open) const redpair

definition
  ordering_constraint_proof :: "'a xmlt \<Rightarrow> 'a redpair xmlt"
where
  "ordering_constraint_proof xml2name = 
     xml1element ''orderingConstraintProof'' (Xml.redpair xml2name) id"
hide (open) const ordering_constraint_proof

definition
  id_labeling :: "('f,'a,'a list)label" 
where
  "id_labeling _ cs = Inl cs" 

definition
  scg_position :: "nat xmlt"
where
  "scg_position \<equiv> Xml.nat ''position''"
hide (open) const scg_position

definition
  scg :: "'f xmlt \<Rightarrow> (('f,string)rule \<times> ((nat \<times> nat)list \<times> (nat \<times> nat)list)) xmlt"
where
  "scg xml2name \<equiv> xml1many_elements ''sizeChangeGraph'' 
    (Xml.rule xml2name)
    (xml3elements ''edge''
       Xml.scg_position
       (Xml.bool ''strict'')
       Xml.scg_position
       (\<lambda>p s q. (s,(p,q)))
    )
    (\<lambda>lr edges. (lr,(map snd (filter fst edges), map snd (filter (\<lambda> x. \<not> fst x) edges))))"
hide (open) const scg

function
  xml2dpProof :: "nat \<Rightarrow> (nat \<Rightarrow> ('f, labelType) lab xmlt) \<Rightarrow> ('f, labelType, string) DPProof xmlt"
where
  "xml2dpProof ld xml2name x = xml1element ''dpProof'' (xml_options [ 
  (''pIsEmpty'',xml0elements ''pIsEmpty'' 
     PisEmpty),
  (''depGraphProc'',xml_many_elements ''depGraphProc''
     (xml2nd_choice ''component'' 
         (xml1element ''dps'' (Xml.rules (xml2name ld)) id)
         ''realScc''
         (xml2dpProof ld xml2name)
         (\<lambda> dps prfOpt. (prfOpt, dps)))
     DepGraphProc),
  (''redPairProc'',xml3elements ''redPairProc'' 
     (Xml.ordering_constraint_proof (xml2name ld)) 
     (xml1element ''dps'' (Xml.rules (xml2name ld)) id) 
     (xml2dpProof ld xml2name) 
     RedPairProc),
  (''redPairUrProc'',xml4elements ''redPairUrProc'' 
     (Xml.ordering_constraint_proof (xml2name ld)) 
     (xml1element ''dps'' (Xml.rules (xml2name ld)) id) 
     (xml1element ''usableRules'' (Xml.rules (xml2name ld)) id) 
     (xml2dpProof ld xml2name)
     RedPairUrProc),
  (''monoRedPairProc'',xml4elements ''monoRedPairProc'' 
     (Xml.ordering_constraint_proof (xml2name ld)) 
     (xml1element ''dps'' (Xml.rules (xml2name ld)) id) 
     (xml1element ''trs'' (Xml.rules (xml2name ld)) id) 
     (xml2dpProof ld xml2name)
     MonoRedPairProc),
  (''monoRedPairUrProc'',xml5elements ''monoRedPairUrProc'' 
     (Xml.ordering_constraint_proof (xml2name ld)) 
     (xml1element ''dps'' (Xml.rules (xml2name ld)) id) 
     (xml1element ''trs'' (Xml.rules (xml2name ld)) id) 
     (xml1element ''usableRules'' (Xml.rules (xml2name ld)) id) 
     (xml2dpProof ld xml2name)
     MonoRedPairUrProc),
  (''semlabProc'',xml4elements ''semlabProc'' 
     (Xml.sl_inter (xml2name ld)) 
     (xml1element ''dps'' (Xml.rules (xml2name (Suc ld))) id) 
     (xml1element ''trs'' (Xml.rules (xml2name (Suc ld))) id) 
     (xml2dpProof (Suc ld) xml2name)
     (\<lambda>i. case i of Some i \<Rightarrow> SemlabProc i id_labeling
                  | None   \<Rightarrow> RootlabProc)),
  (''unlabProc'',xml3elements ''unlabProc''
     (xml1element ''dps'' (Xml.rules (xml2name (ld - 1))) id) 
     (xml1element ''trs'' (Xml.rules (xml2name (ld - 1))) id) 
     (xml2dpProof (ld - 1) xml2name)
     UnlabProc),
  (''subtermProc'',xml1many2elements_gen ''subtermProc''
     (Xml.sp (xml2name ld))
     (\<lambda>\<pi>. Xml.projected_rseq (xml2name ld) (create_sp \<pi>))
     (xml1element ''dps'' (Xml.rules (xml2name ld)) id)
     (xml2dpProof ld xml2name)
     SubtermCriterionProc),
  (''flatContextClosureProc'', xml5elements ''flatContextClosureProc''
    (xml1element ''freshSymbol'' (xml2name ld) id)
    (Xml.flat_contexts (xml2name ld))
    (xml1element ''dps'' (Xml.rules (xml2name ld)) id)
    (xml1element ''trs'' (Xml.rules (xml2name ld)) id)
    (xml2dpProof ld xml2name)
    FccProc),
  (''sizeChangeProc'', xml1many_elements ''sizeChangeProc''
     (xml_options [
         (''subtermCriterion'', xml0elements ''subtermCriterion'' (None)),
         (''reductionPair'', xml2elements ''reductionPair'' 
              (Xml.ordering_constraint_proof (xml2name ld)) 
              (xml1element ''usableRules'' (Xml.rules (xml2name ld)) id) 
              (\<lambda>redp ur. Some (redp,ur)))
     ])
     (Xml.scg (xml2name ld))
     (\<lambda>version. case version of 
         None \<Rightarrow> SizeChangeSubTerm
       | Some redp_ur \<Rightarrow> SizeChangeRedPair (fst redp_ur) (snd redp_ur))
  )    
  ]) id x"
by (pat_completeness) auto

definition
  xmlIgnore :: "string \<Rightarrow> unit xmlt"
where
  "xmlIgnore name xml \<equiv> if name = (Xml.get_name xml) then return () else error (''expected '' @ name @ '' element, but found '' @ to_string xml)"

function
  xml2trsProof ::
    "('f, labelType) lab xmlt \<Rightarrow> ('f, labelType) lab shp xmlt
     \<Rightarrow> (nat \<Rightarrow> (('f, labelType) lab shp, labelType)lab xmlt)
     \<Rightarrow> ('f, labelType, string) TRSProof xmlt"
where "xml2trsProof xml2name xml2sname xml2lsname x = xml1element ''trsTerminationProof'' (xml_options [
   (''rIsEmpty'',xml0elements ''rIsEmpty''
      RisEmpty),
   (''dpTrans'',xml3elements ''dpTrans'' 
      (xml1element ''dps'' (Xml.rules xml2sname) id)
      (Xml.bool ''markedSymbols'')
      (xml2dpProof 0 xml2lsname)
      (\<lambda> dps _ prf. DPTrans dps prf)),
   (''ruleRemoval'',xml3elements ''ruleRemoval''
      (Xml.ordering_constraint_proof xml2name) 
      (xml1element ''trs'' (Xml.rules xml2name) id)
      (xml2trsProof xml2name xml2sname xml2lsname)
      RuleRemoval),
  (''unlab'',xml2elements ''unlab''
     (xml1element ''trs'' (Xml.rules xml2name) id) 
     (xml2trsProof xml2name xml2sname xml2lsname)
     Unlab),
  (''stringReversal'',xml2elements ''stringReversal''
     (xml1element ''trs'' (Xml.rules xml2name) id) 
     (xml2trsProof xml2name xml2sname xml2lsname)
     StringReversal),
  (''semlab'',xml3elements ''semlab'' 
     (Xml.sl_inter xml2name) 
     (xml1element ''trs'' (Xml.rules xml2name) id) 
     (xml2trsProof xml2name xml2sname xml2lsname)
     (\<lambda>i. case i of Some i \<Rightarrow> Semlab i id_labeling
                  | None   \<Rightarrow> Rootlab)),
  (''flatContextClosure'',xml3elements ''flatContextClosure''
    (Xml.flat_contexts xml2name)
    (xml1element ''trs'' (Xml.rules xml2name) id)
    (xml2trsProof xml2name xml2sname xml2lsname)
    Fcc
  )
  ]) id x"
by (pat_completeness, auto)

function
  xml2dpDisproof :: "'f xmlt \<Rightarrow> ('f, string) DPDisproof xmlt"
where
  "xml2dpDisproof xml2name = xml1element ''dpNonterminationProof'' (xml_options [
   (''dpRuleRemoval'',xml3elements ''dpRuleRemoval''
      (xml1element ''dps'' (Xml.rules xml2name) id)
      (xml1element ''trs'' (Xml.rules xml2name) id)
      (xml2dpDisproof xml2name)
      DPRuleRemoval),
   (''loop'', xml_change (Xml.loop xml2name)
        (\<lambda> (s, rseq, \<sigma>, C). DPLoop s rseq \<sigma> C))
   ]) id"
by pat_completeness auto

function
  xml2trsDisproof :: "'f xmlt \<Rightarrow> ('f shp) xmlt \<Rightarrow> ('f, string) TRSDisproof xmlt"
where
  "xml2trsDisproof xml2name xml2sname = xml1element ''trsNonterminationProof'' (xml_options [
   (''variableConditionViolated'',xml0elements ''variableConditionViolated''
      NotWellFormed),
   (''loop'', xml_change (Xml.loop xml2name)
        (\<lambda> (s, rseq, \<sigma>, C). Loop s (map (\<lambda> (x,y,_,z). (x,y,z)) rseq) \<sigma> C)),
   (''ruleRemoval'',xml2elements ''ruleRemoval''
      (xml1element ''trs'' (Xml.rules xml2name) id)
      (xml2trsDisproof xml2name xml2sname)
      NTRuleRemoval),
   (''dpTrans'',xml3elements ''dpTrans''
      (xml1element ''dps'' (Xml.rules xml2sname) id)
      (Xml.bool ''markedSymbols'')
      (xml2dpDisproof xml2sname)
      (\<lambda> p _ prf. NTDPTrans p prf)),
   (''stringReversal'',xml2elements ''stringReversal''
      (xml1element ''trs'' (Xml.rules xml2name) id)
      (xml2trsDisproof xml2name xml2sname)
      NTStringReversal)
   ]) id"
by pat_completeness auto

definition
  xml2trsInput :: "'f xmlt \<Rightarrow> ('f,string) rules xmlt"
where
  "xml2trsInput xml2name \<equiv> xml1element ''input''
   (xml1element ''trsInput''
      (xml1element ''trs'' (Xml.rules xml2name) id)
      id)
   id"

fun
  xml2certProblem :: "
  'f xmlt \<Rightarrow> 
  ('f shp) xmlt \<Rightarrow> 
  ('f,labelType)lab xmlt \<Rightarrow> 
  ('f,labelType)lab shp xmlt \<Rightarrow> 
  (nat \<Rightarrow> (('f, labelType) lab shp, labelType) lab xmlt) \<Rightarrow> 
  ('f, labelType, string) certificationProblem xmlt"
where "xml2certProblem xml2uname xml2usname xml2name xml2sname xml2lsname (XML name egal cs text) = (if name = ''certificationProblem'' \<and> text = None \<and> length cs \<ge> 3 then 
  xml2trsInput xml2uname (cs ! 0) 
      \<guillemotright>= (\<lambda> trs. (Xml.text ''cpfVersion'') (cs ! 1)
      \<guillemotright>= (\<lambda> version. if (take 2 version = ''2.'') then 
         xml1element ''proof'' (
            xml_options [
              (''trsTerminationProof'', xml_change (xml2trsProof xml2name xml2sname xml2lsname) (\<lambda> p. TRSProof trs p)),
              (''trsNonterminationProof'', xml_change (xml2trsDisproof xml2uname xml2usname) (\<lambda> p. TRSDisproof trs p))
                       ]
         ) id (cs ! 2)
      else error ''cpf-version 2.x required''))
  else error ''could not parse root element'')"

primrec
  xmldoc2certProblem :: "xmldoc \<Rightarrow> string + (string,labelType,string)certificationProblem"
where
  "xmldoc2certProblem (XMLDOC header xml) = debug ''0'' ''parsing xml to certification problem'' (
     xml2certProblem (Xml.text ''name'') Xml.sharp_name Xml.plain_name Xml.sharp_lab_name Xml.lab_sharp_name xml)"

fun
  parseCertProblem :: "string \<Rightarrow> string + (string,labelType,string)certificationProblem"
where
  "parseCertProblem s = (doc_of_string s \<guillemotright>= xmldoc2certProblem)"

(*
fun parseTRSProof :: "string \<Rightarrow> (string,labelType,string)TRSProof"
where "parseTRSProof s = (case parseCertProblem s of
  Left err \<Rightarrow> undefined
| Right cp \<Rightarrow> (case cp of
    CertificationProblem term trs (Left trsproof) \<Rightarrow> trsproof
  | _ \<Rightarrow> undefined
  )
)"
*)

declare parseCertProblem.simps[simp del]

(* ... to xml *)

class xml =  
  fixes xml :: "'a \<Rightarrow> xml"

instantiation "term" :: ("show", "show") xml
begin
fun xml_term 
  where "xml_term (Var x) = XML ''var'' [] [] (Some (to_string x))" 
      | "xml_term (Fun f ts) = XML ''funapp'' [] 
              (XML ''name'' [] [] (Some (to_string f)) # 
              (map (\<lambda> t. XML ''arg'' [] [xml_term t] None) ts)) 
           None"
instance ..
end


instantiation certificationProblem :: ("show", type, "show") xml
begin
fun xml_rules :: "('a,'c) rules \<Rightarrow> xml"
where "xml_rules rls = XML ''trs'' Nil [XML ''rules'' [] (map (\<lambda> (l,r). XML ''rule'' [] [
         XML ''lhs'' [] [xml l] None,         
         XML ''rhs'' [] [xml r] None
         ] None) rls) None] None"
fun xml_certificationProblem :: "('a,'b,'c)certificationProblem \<Rightarrow> xml"
where "xml_certificationProblem cp = (let rls = (case cp of TRSProof lr _ \<Rightarrow> lr | TRSDisproof lr _ \<Rightarrow> lr) in
     (XML ''trsInput'' Nil 
        [xml_rules rls]
        None))"
instance ..
end

end
